//---------------------------------------------------------------------------
// Copyright (C) 1999 by Dmitriy Vassiliev
// SlyDiman@mail.ru
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "DDForm.h"
#include <designintf.hpp>
#include <designeditors.hpp>

#include "DDSurfac.h"
#include "DDUtil.h"
#include "GetDXVer.h"

#pragma package(smart_init)
#pragma link "DDUtil"
#pragma link "AlphaBlt"
#pragma link "TileBlt"
#pragma link "GetDXVer"

#pragma resource "*.res"

//---------------------------------------------------------------------------
namespace Ddform
{
    void __fastcall PACKAGE Register()
    {
        RegisterCustomModule( __classid(TDDForm), __classid(TCustomModule) );
    }
}

//---------------------------- TDDSurfaceList -------------------------

#define SListIndexError   "List index out of bounds (%d)"

TDDSurface *  __fastcall TDDSurfaceList::Get(int Index)
{
  if( Index < 0 ||  Index >= Count )  Error( __classid(TDDSurfaceList), SListIndexError, Index );
  return( (TDDSurface *)List[Index] );
}

void __fastcall TDDSurfaceList::Put( int Index, TDDSurface * Item)
{
  if( Index < 0 || Index >= Count )  Error( __classid(TDDSurfaceList), SListIndexError, Index );
  List[Index] = (void *)Item;
}

void __fastcall TDDSurfaceList::ClearItems(void)
{
  for( int i=0; i < Count; i++ ){
    delete Items[i];
  }
  Clear();
}


//---------------------------------------------------------------------------
__fastcall TDDForm::TDDForm(TComponent* Owner) : TForm(CreateInit(Owner))
{
  if( ComponentState.Contains(csDesigning) ){
    MessageBox(NULL, "This is a DirectDraw Form v2.0 by Dmitriy Vassiliev",
                     "Create DirectDraw Form", MB_OK | MB_ICONINFORMATION );
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::EnumSurfaces(void)
{
  Surfaces->Clear();
  for( int i=0; i < ComponentCount; i++ ){
    if( Components[i]->ClassNameIs("TDDSurface") ){
//      Surfaces->Add( (TDDSurface *)Components[i] );
      Surfaces->Insert( 0, (TDDSurface *)Components[i] );
    }
  }

}

//---------------------------------------------------------------------------
void __fastcall TDDForm::Loaded(void)
{
  TForm::Loaded();

  strcpy( _PalFileName, FPalFileName.c_str() );

  Surfaces = new TDDSurfaceList();

  EnumSurfaces();

  if( ComponentState.Contains(csDesigning) )  return;

  if( DDWindowed ){
    BorderStyle = bsSizeable;
  }else{
    BorderStyle = bsNone;
  }

  DWORD dwDirectXVersion = 0;
  char strDirectXVersion[10];
  GetDXVersion( &dwDirectXVersion, strDirectXVersion, 10 );

  if( dwDirectXVersion < 0x7000 ){
    throw Exception("TDDForm: Need DirectX 7.0 or better");
  }

  hDDraw = LoadLibrary("DDRAW.DLL");
  if( hDDraw ){
    pDirectDrawCreate = (TDirectDrawCreate)GetProcAddress( hDDraw, "DirectDrawCreate" );
    if( pDirectDrawCreate == NULL ){
      FreeLibrary( hDDraw );
      hDDraw = NULL;
    }
  }

  if( pDirectDrawCreate == NULL || pDirectDrawCreate( NULL, &pDDraw1, NULL) != DD_OK )
    throw Exception("TDDForm: Create DirectDraw Object FAILED");

  if( pDDraw1->QueryInterface( IID_IDirectDraw7, (LPVOID*)&pDDraw) != DD_OK )
    throw Exception("TDDForm: QueryInterface FAILED");

  Application->OnException = ExceptionHandler;

  // InitSurfaces();
}

//---------------------------------------------------------------------------
TComponent* __fastcall TDDForm::CreateInit(TComponent* AOwner)
{
  FDDWidth = 640;
  FDDHeight = 480;
  FDDBPP = 16;
  FAppPaused = false;

  PrimaryCanvas = NULL;
  BackCanvas = NULL;

  pDDraw1 = NULL;       // DirectDraw object
  pDDraw = NULL;        // DirectDraw object
  pDDPrimary = NULL;    // DirectDraw primary surface
  pDDBack = NULL;       // DirectDraw back surface
  pDDPalette = NULL;    // The primary surface palette
  pDDClipper = NULL;    // Clipper for primary

  DDThread = NULL;

  FDDActive = false;        // App is running/active (not hidden)
  DDHidden = false;
  DDReady = false;         // App is ready for updates
  DDMenu = false;

  Ticks = 0;
  TicksPause = 0;

  AlphaEx = NULL;

  return( AOwner );
}

//---------------------------------------------------------------------------
__fastcall TDDForm::~TDDForm(void)
{
  DDActive = false;

  Application->OnException = NULL;

  if( pDDraw  )  pDDraw->Release();
  if( pDDraw1 )  pDDraw1->Release();

  delete Surfaces;

  if( hDDraw ){
    FreeLibrary( hDDraw );
    hDDraw = NULL;
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::WndProc( Messages::TMessage &Message )
{

  switch( Message.Msg ){
  case WM_ENTERMENULOOP:
    if( DDReady ){
      DDSCAPS2 caps;
      pDDPrimary->GetCaps(&caps);
      if( caps.dwCaps & DDSCAPS_MODEX )  ToggleScreen( DDF_TOGGLE_WINDOWED );
      pDDraw->FlipToGDISurface();
    }
    DrawMenuBar(Handle);
    DDMenu = true;
    break;

  case WM_EXITMENULOOP:
    DDMenu = false;
    break;

  case WM_SIZE:
    {
      switch( Message.WParam ){
      case SIZE_RESTORED:
        DDHidden = false;
        ToggleScreen( DDF_TOGGLE_WINDOWED );
        break;
      case SIZE_MAXSHOW:
      case SIZE_MAXIMIZED:
        DDHidden = false;
        ToggleScreen( DDF_TOGGLE_FULLSCREEN );
        break;
      case SIZE_MAXHIDE:
      case SIZE_MINIMIZED:
        DDHidden = true;
        break;
      }
      break;
    }

  case WM_MOVE:
    {
      // Retrieve the window position after a move
      if( DDActive && DDReady && DDWindowed && !DDHidden){
        ::GetWindowRect(Handle, &rcWindow );
      	::GetClientRect(Handle, &rcViewport );
    	::GetClientRect(Handle, &rcScreen );
    	::ClientToScreen(Handle, (POINT*)&(rcScreen.left) );
    	::ClientToScreen(Handle, (POINT*)&(rcScreen.right) );
      }
      break;
    }

  case WM_PALETTECHANGED:
    if( (HWND)Message.WParam == Handle )  break;

  case WM_QUERYNEWPALETTE:
    if( DDReady ){
      if( pDDPrimary && pDDPalette )  pDDPrimary->SetPalette( pDDPalette );
      UpdateSurfaces();
      if( OnDDChangePalette )  OnDDChangePalette(this);
    }
    return;

  }

  TForm::WndProc(Message);

  switch( Message.Msg ){
  case WM_PAINT:
    if( DDWindowed && DDReady ){
      while( true )
        // If we are in windowed mode, perform a blt.
        if( MakeItSo( pDDPrimary->Blt(&rcScreen, pDDBack,
                                     &rcViewport, DDBLT_WAIT, NULL)) ) break;
    }
    break;
  }

}

//---------------------------------------------------------------------------
void __fastcall TDDForm::ReleaseAllObjects(void)
{
  DDReady = false;

  if( AlphaEx ){
    delete AlphaEx;
    AlphaEx = NULL;
  }  

  if( pDDraw ){
    pDDraw->SetCooperativeLevel( Handle, DDSCL_NORMAL );

    delete BackCanvas;
    BackCanvas = NULL;

    delete PrimaryCanvas;
    PrimaryCanvas = NULL;

    if( pDDBack ){
      pDDBack->Release();
      pDDBack = NULL;
    }

    if( pDDPrimary ){
      pDDPrimary->Release();
      pDDPrimary = NULL;
    }

    if( pDDClipper ){
      pDDClipper->Release();
      pDDClipper = NULL;
    }

    for( int i=0; i < Surfaces->Count; i++ ){
      Surfaces->Items[i]->Release();
    }

    if( pDDPalette ){
      pDDPalette->Release();
      pDDPalette = NULL;
    }
  }

}

//---------------------------------------------------------------------------
void __fastcall TDDForm::ExceptionHandler(TObject *Sender, Exception *E)
{
  DDActive = false;

  Application->MessageBox( E->Message.c_str(), "DirectDraw fatal error", MB_OK | MB_ICONERROR );
  Application->Terminate();
}

//-----------------------------------------------------------------------------
void __fastcall TDDForm::InitSurfaces(void)
{
  DDSCAPS2            ddscaps;

  DDReady = false;

  for( int i=0; i < Surfaces->Count; i++ )  Surfaces->Items[i]->Release();
  EnumSurfaces();

  if( DDWindowed ){
    // Get normal windowed mode
    if( pDDraw->SetCooperativeLevel(Handle, DDSCL_NORMAL) != DD_OK )
      throw Exception("TDDForm: Set Cooperative Level FAILED");

  	// Get the dimensions of the viewport and screen bounds
   	::GetClientRect(Handle, &rcViewport);
   	::GetClientRect(Handle, &rcScreen);
   	::ClientToScreen(Handle, (POINT*)&rcScreen.left);
   	::ClientToScreen(Handle, (POINT*)&rcScreen.right);

    // Create the primary surface
    ZeroMemory(&PrimaryDesc,sizeof(PrimaryDesc));
    PrimaryDesc.dwSize = sizeof(PrimaryDesc);
    PrimaryDesc.dwFlags = DDSD_CAPS;
    PrimaryDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    if( pDDraw->CreateSurface(&PrimaryDesc, &pDDPrimary, NULL) != DD_OK )
      throw Exception("TDDForm: Create Primary Surface FAILED (Win)");

    // Create a clipper object since this is for a Windowed render
    LPDIRECTDRAWCLIPPER pClipper;
    if( pDDraw->CreateClipper(0, &pClipper, NULL) != DD_OK)
      throw Exception("TDDForm: Create Clipper FAILED (Win)");

    // Associate the clipper with the window
    pClipper->SetHWnd( 0, Handle );
    pDDPrimary->SetClipper(pClipper);
    pClipper->Release();
    pClipper = NULL;

    // Get the backbuffer. For fullscreen mode, the backbuffer was created
    // along with the primary, but windowed mode still needs to create one.
    ZeroMemory(&BackDesc,sizeof(BackDesc));
    BackDesc.dwSize = sizeof(BackDesc);
    BackDesc.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
    BackDesc.dwWidth        = DDWidth;
    BackDesc.dwHeight       = DDHeight;
    BackDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    if( pDDraw->CreateSurface(&BackDesc, &pDDBack, NULL) != DD_OK)
      throw Exception("TDDForm: Create Back Surface FAILED (Win)");

  }else{
     // Get exclusive mode
     if( pDDraw->SetCooperativeLevel(Handle, DDSCL_EXCLUSIVE |
                                           DDSCL_ALLOWREBOOT |
                                           DDSCL_ALLOWMODEX |
                                           DDSCL_FULLSCREEN) != DD_OK)
       throw Exception("TDDForm: Set Cooperative Level FAILED (FS)");

     // Set the video mode
     if( pDDraw->SetDisplayMode( DDWidth, DDHeight, DDBPP, 0, 0) != DD_OK)
       throw Exception("TDDForm: Set Display Mode FAILED (FS)");

  	// Get the dimensions of the viewport and screen bounds
   	// Store the rectangle which contains the renderer
   	 SetRect(&rcViewport, 0, 0, DDWidth, DDHeight );
   	 memcpy(&rcScreen, &rcViewport, sizeof(RECT) );

    // Create the primary surface with 1 back buffer
    ZeroMemory(&PrimaryDesc,sizeof(PrimaryDesc));
    PrimaryDesc.dwSize = sizeof(PrimaryDesc);
    PrimaryDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    PrimaryDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
              		         DDSCAPS_FLIP |
               	                 DDSCAPS_COMPLEX; // | DDSCAPS_VIDEOMEMORY;
    PrimaryDesc.dwBackBufferCount = 1;
    if( pDDraw->CreateSurface( &PrimaryDesc, &pDDPrimary, NULL) != DD_OK ){
//      PrimaryDesc.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
//      if( pDDraw->CreateSurface( &PrimaryDesc, &pDDPrimary, NULL) != DD_OK )
        throw Exception("TDDForm: Create Primary Surface FAILED (FS)");
    }

    // Create a clipper object since this is for a Windowed render
    if( pDDraw->CreateClipper(0, &pDDClipper, NULL) != DD_OK)
      throw Exception("TDDForm: Create Clipper FAILED (FS)");

    // Associate the clipper with the window
    pDDClipper->SetHWnd( 0, Handle );
    FixClip();

    // Get a pointer to the back buffer
    ZeroMemory(&ddscaps, sizeof(ddscaps));
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    if( pDDPrimary->GetAttachedSurface(&ddscaps, &pDDBack) != DD_OK)
      throw Exception("TDDForm: Get Attached Surface FAILED (FS)");

    ZeroMemory(&BackDesc, sizeof(BackDesc));
    BackDesc.dwSize = sizeof(BackDesc);
    if( pDDBack->GetSurfaceDesc(&BackDesc) != DD_OK )
      throw Exception("Get Back Surface Desc FAILED");
  }

//------------------------------------ Black screen
  {
    DDBLTFX     ddbltfx;

    // Use the blter to do a color fill to clear the back buffer
    ZeroMemory(&ddbltfx, sizeof(ddbltfx));
    ddbltfx.dwSize = sizeof(ddbltfx);
    ddbltfx.dwFillColor = 0;
    pDDPrimary->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
    pDDBack->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
  }

//------------------------------------
  PrimaryCanvas = new TDDCanvas( pDDPrimary );
  BackCanvas = new TDDCanvas( pDDBack );

//------------------------------------ Loading...
  try{
    BackCanvas->Font->Size = 14;
    BackCanvas->Font->Color = (TColor)0xFFFFFF;
    BackCanvas->Brush->Style = bsClear;
    AnsiString s = "Loading...";
    int w = BackCanvas->TextWidth(s);
    BackCanvas->TextOut( DDWidth/2 - w/2, DDHeight/2 - 16, s );
  }
  __finally{
    BackCanvas->Release();
  }


//------------------------------------
  if( AlphaEx )  delete AlphaEx;
  AlphaEx = NewAlphaBlitter( pDDPrimary );
//  if( AlphaEx == NULL )  throw Exception("Can't create Alpha Blitter object");

//------------------------------------
  pDDPalette = NULL;

  DDSURFACEDESC2 ddsd;
  ZeroMemory(&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);
  if( pDDPrimary->GetSurfaceDesc(&ddsd) != DD_OK )
      throw Exception("TDDForm: Get Primary Surface Desc FAILED");

  if( ddsd.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8 ){
    // create a palette if we are in a paletized display mode.
    //
    // NOTE because we want to be able to show dialog boxs and
    // use our menu, we leave the windows reserved colors as is
    // so things dont look ugly.
    //
    // palette is setup like so:
    //
    //      10      windows system colors
    //      64      red wash
    //      64      grn wash
    //      64      blu wash

    if( _PalFileName[0] == 0 ){
      // get the current windows colors.
      PALETTEENTRY ape[256];
      HDC hdc = GetDC(NULL);
      GetSystemPaletteEntries(hdc, 0, 256, ape);
      ReleaseDC(NULL, hdc);

      for( int i=0; i<64; i++ ){
        ape[10+64*0+i].peRed   = i * 255/63;
        ape[10+64*0+i].peGreen = 0;
        ape[10+64*0+i].peBlue  = 0;

        ape[10+64*1+i].peRed   = 0;
        ape[10+64*1+i].peGreen = i * 255/63;
        ape[10+64*1+i].peBlue  = 0;

        ape[10+64*2+i].peRed   = 0;
        ape[10+64*2+i].peGreen = 0;
        ape[10+64*2+i].peBlue  = i * 255/63;
      }
/*
      for (i = 0; i < 10; i++)
        ape[i].peFlags = D3DPAL_READONLY;
      for (i = 10; i < 246; i++)
        ape[i].peFlags = D3DPAL_FREE | PC_RESERVED;
      for (i = 246; i < 256; i++)
        ape[i].peFlags = D3DPAL_READONLY;
*/
      // create the palette.
      if( pDDraw->CreatePalette(DDPCAPS_8BIT, ape, &pDDPalette, NULL) ==  DD_OK ){
        if( pDDPrimary->SetPalette( pDDPalette ) != DD_OK ) ;
//      throw Exception("TDDForm: Set internal Palette FAILED");
      }
    }else{
      pDDPalette = DDLoadPalette( pDDraw, _PalFileName );
      if( pDDPalette ){
        if( pDDPrimary->SetPalette( pDDPalette ) != DD_OK ) ;
//      throw Exception("TDDForm: Set Palette from file FAILED");
      }
    }
  }

//------------------------------------
  for( int i=0; i < Surfaces->Count; i++ )  Surfaces->Items[i]->Init();

  if( OnDDChangeSize    )  OnDDChangeSize(this);
  if( OnDDChangePalette )  OnDDChangePalette(this);

  DDReady = true;
}


//---------------------------------------------------------------------------
void __fastcall TDDForm::SetFullScreenClip(bool c)
{
  if( FFullScreenClip == c )  return;
  FFullScreenClip = c;

  if( DDWindowed )  return;
  FixClip();
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::FixClip(void)
{
  if( DDFullScreenClip ){
    pDDraw->FlipToGDISurface();
    pDDPrimary->SetClipper( pDDClipper );
  }else{
    pDDPrimary->SetClipper( NULL );
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::ToggleScreen(int mode)
{
  if( mode == DDF_TOGGLE_WINDOWED && DDWindowed )  return;
  if( mode == DDF_TOGGLE_FULLSCREEN && !DDWindowed )  return;

  if( !(DDActive && DDReady && !DDHidden) )  return;

  DDActive = false;

  if( DDWindowed )  GetWindowRect(Handle, &rcWindow);

  DDWindowed = !DDWindowed;

  if( DDWindowed ){
    BorderStyle = bsSizeable;
  }else{
    BorderStyle = bsNone;
  }

  // In case we're coming from a fullscreen mode, restore the window size
  if( DDWindowed ){
    SetWindowPos(Handle, HWND_NOTOPMOST, rcWindow.left, rcWindow.top,
                (rcWindow.right - rcWindow.left),
                (rcWindow.bottom - rcWindow.top), SWP_SHOWWINDOW );
  }

  // Re-create the surfaces
  DDActive = true;
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::FixSize(void)
{
  if( ComponentState.Contains(csDesigning) )  return;

  Constraints->MinWidth = 0;
  Constraints->MaxWidth = 0;
  Constraints->MinHeight = 0;
  Constraints->MaxHeight = 0;

  if( !DDActive )  return;

  if( DDWindowed ){
    ClientWidth  = DDWidth;
    ClientHeight = DDHeight;
  }else{
    Width  = DDWidth;
    Height = DDHeight;
  }

  Constraints->MinWidth = Width;
  Constraints->MaxWidth = Width;
  Constraints->MinHeight = Height;
  Constraints->MaxHeight = Height;
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::SetDDWidth(int x)
{
  FDDWidth = x;
  FixSize();
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::SetDDHeight(int y)
{
  FDDHeight = y;
  FixSize();
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::SetDDBPP(int b)
{
  FDDBPP = b;
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::ChangeSize(int x, int y, int bpp)
{
  if( !DDReady ){
    FDDWidth = x;
    FDDHeight = y;
    FDDBPP = bpp;
  }else{
    bool tmp = DDActive;
    DDActive = false;
    if( DDWindowed )  GetWindowRect(Handle, &rcWindow);

    FDDWidth = x;
    FDDHeight = y;
    FDDBPP = bpp;
    FixSize();

    DDActive = tmp;
  }
}

//---------------------------------------------------------------------------
bool __fastcall TDDForm::MakeItSo( HResult DDResult )
{
  if( DDResult == DD_OK ){
    return( true );
  }else if( DDResult == DDERR_SURFACELOST ){
    return( !RestoreAll() );
  }else{
    return( DDResult != DDERR_WASSTILLDRAWING );
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::SetAppPaused(bool _AppPaused)
{
  if( FAppPaused == _AppPaused )  return;

  if( FAppPaused ){
    TicksPause = GetTickCount() - TicksPause;
    FAppPaused = false;
  }else{
    FAppPaused = true;
    TicksPause = GetTickCount();
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::SetDDActive(bool _DDActive)
{
  if( FDDActive == _DDActive )  return;
  if( FDDActive ){
    if( DDThread ){
      DDThread->Terminate();
      DDThread->WaitFor();
      delete DDThread;
      DDThread = NULL;
    }
    FDDActive = false;
    if( OnDDActiveChange )  OnDDActiveChange(this);
    ReleaseAllObjects();
    FixSize();
  }else{
    FDDActive = true;
    FixSize();
    InitSurfaces();
    DDThread = new TDDThread(this);
    DDThread->Priority = tpNormal;  // tpHigher, tpHighest
    DDThread->Resume();

    if( OnDDActiveChange )  OnDDActiveChange(this);
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::UpdateFrame(void)
{
  if( DDPaused || DDMenu || !(DDActive && DDReady && !DDHidden) )  return;

  Ticks = GetTickCount() + TicksPause;

  if( OnDDUpdateDisplay )  OnDDUpdateDisplay(this);

  while( true )
    if( DDWindowed ){
      // If we are in windowed mode, perform a blt.
      // BltFast not work with attached clippers
      if( MakeItSo(pDDPrimary->Blt(&rcScreen, pDDBack,
                                   &rcViewport, DDBLT_WAIT, NULL)) ) break;

    }else{
      // Else we are in fullscreen mode, so perform a flip.

      if( DDFullScreenClip ){
        if( MakeItSo(pDDPrimary->Blt(NULL, pDDBack,
                                     NULL, DDBLT_WAIT, NULL)) ) break;
      }else{
        if( MakeItSo(pDDPrimary->Flip( NULL, 0L )) )  break;
      }
    }

/*
  bool Drawing = false;
  for( int i=0; i < Surfaces->Count; i++ ){
    if( Surfaces->Items[i]->pDDCanvas->DrawingAllowed() ){
      Drawing = true;
      break;
    }
  }
  Done = !Drawing;
*/
}

//---------------------------------------------------------------------------
bool __fastcall TDDForm::RestoreAll( void )
{
  HResult Result;

  Result = pDDPrimary->Restore();
  if( Result != DD_OK )    return( false );

  Result = pDDBack->Restore();
  if( Result != DD_OK )    return( false );

  for( int i=0; i < Surfaces->Count; i++ ){
    Result = Surfaces->Items[i]->pDDSurface->Restore();
    if( Result != DD_OK )    return( false );
  }

  UpdateSurfaces();
  return( true );
}

//---------------------------------------------------------------------------
void __fastcall TDDForm::UpdateSurfaces( void )
{
  for( int i=0; i < Surfaces->Count; i++ )  Surfaces->Items[i]->Update();
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall TDDThread::Execute(void)
{
  while( true ){
    if( Terminated )  break;
    DDForm->UpdateFrame();
  }
}

//---------------------------------------------------------------------------

//------------------------- Examples ----------------------------------------

//---------------------------------------------------------------------------

/*
//                     Color fill example

    DDBLTFX     ddbltfx;
    HDC         hDC;

    // Use the blter to do a color fill to clear the back buffer
    ZeroMemory(&ddbltfx, sizeof(ddbltfx));
    ddbltfx.dwSize = sizeof(ddbltfx);
    ddbltfx.dwFillColor = 0;
    g_pDDSBack->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);

    if (g_pDDSBack->GetDC(&hDC) == DD_OK)
    {
        // ...
        g_pDDSBack->ReleaseDC(hDC);
    }
*/

//---------------------------------------------------------------------------

/*
//                       Toggle Pause

    DDSCAPS2 caps;
    pDDPrimary->GetCaps(&caps);
    if( caps.dwCaps & DDSCAPS_MODEX )  ToggleScreen( DDF_TOGGLE_WINDOWED );
    pDDraw->FlipToGDISurface();
    DrawMenuBar(Handle);
//  RedrawWindow(Handle, NULL, NULL, RDW_FRAME);
    Repaint();
*/

//---------------------------------------------------------------------------
