//---------------------------------------------------------------------------
// Copyright (C) 1999 by Dmitriy Vassiliev
// SlyDiman@mail.ru
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#ifndef DDFormH
#define DDFormH

//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Forms.hpp>

#define   DIRECTDRAW_VERSION 0x0700
#include <DDraw.h>

#include "DDCanvas.h"
#include "AlphaBlt.h"
#include "TileBlt.h"

//---------------------------------------------------------------------------

#define  DDF_TOGGLE_AUTO          (0)
#define  DDF_TOGGLE_WINDOWED      (1)
#define  DDF_TOGGLE_FULLSCREEN    (2)

//---------------------------------------------------------------------------
class TDDSurface;

typedef TDDSurface * TDDSurfaceListItem[134217727];
typedef TDDSurfaceListItem * PDDSurfaceListItem;

class PACKAGE TDDSurfaceList : public TList
{
protected:
	TDDSurface * __fastcall Get(int Index);
	void __fastcall Put(int Index, TDDSurface * Item);

public:
  void _fastcall ClearItems(void);
  __property TDDSurface * Items[int Index] = {read=Get, write=Put};
  __fastcall TDDSurfaceList( void ) : TList() {}
};

class TDDThread;

//---------------------------------------------------------------------------
typedef HRESULT  __stdcall (*TDirectDrawCreate)( GUID FAR *lpGUID, LPDIRECTDRAW FAR *lplpDD, IUnknown FAR *pUnkOuter );

//---------------------------------------------------------------------------
class PACKAGE TDDForm : public TForm
{
private:
    TDirectDrawCreate pDirectDrawCreate;
    HMODULE hDDraw;

    TDDSurfaceList *Surfaces;

    TDDThread *DDThread;

    System::AnsiString FPalFileName;
    char _PalFileName[256];
    bool FAppWindowed;
    bool FAppPaused;
    bool FFullScreenClip;
    bool FDDActive;
    int FDDWidth;
    int FDDHeight;
    int FDDBPP;
    int AppIdleTick;
    TNotifyEvent FOnActiveChange;
    TNotifyEvent FOnUpdateDisplay;
    TNotifyEvent FOnChangeSize;
    TNotifyEvent FOnChangePalette;
    unsigned FColorKeyRGB;

    void __fastcall ExceptionHandler(TObject *Sender, Exception *E);
    bool __fastcall RestoreAll( void );
    void __fastcall IdleHandler( TObject *Sender, bool &Done ) ;
    void __fastcall InitSurfaces(void);
    void __fastcall ReleaseAllObjects(void);
    void __fastcall SetAppPaused(bool _AppPaused);
    void __fastcall SetDDActive(bool _DDActive);
    void __fastcall SetFullScreenClip(bool c);
    void __fastcall SetDDWidth(int);
    void __fastcall SetDDHeight(int);
    void __fastcall SetDDBPP(int);
    void __fastcall FixSize(void);
    void __fastcall FixClip(void);

    TComponent* __fastcall TDDForm::CreateInit(TComponent* AOwner);

protected:
    virtual void __fastcall Loaded(void);
    virtual void __fastcall WndProc(Messages::TMessage &Message);

public:
    TDDCanvas *PrimaryCanvas;
    TDDCanvas *BackCanvas;

    LPDIRECTDRAW  pDDraw1;
    LPDIRECTDRAW7 pDDraw;              // DirectDraw object
    LPDIRECTDRAWSURFACE7 pDDPrimary;   // DirectDraw primary surface
    LPDIRECTDRAWSURFACE7 pDDBack;      // DirectDraw back surface
    LPDIRECTDRAWPALETTE  pDDPalette;   // The primary surface palette
    LPDIRECTDRAWCLIPPER  pDDClipper;

    DDSURFACEDESC2 PrimaryDesc;
    DDSURFACEDESC2 BackDesc;

    CAlphaBlitter *AlphaEx;

    RECT rcWindow;        // Saves the window size & pos.
    RECT rcViewport;      // Pos. & size to blt from
    RECT rcScreen;        // Screen pos. for blt
    bool DDHidden;
    bool DDReady;         // App is ready for updates
    bool DDMenu;
    DWORD Ticks;
    DWORD TicksPause;

    void __fastcall EnumSurfaces(void);
    void __fastcall UpdateFrame(void);
    void __fastcall UpdateSurfaces( void );
    bool __fastcall MakeItSo( System::HResult DDResult );
    void __fastcall ToggleScreen(int DDF_TOGGLE);
    void __fastcall ChangeSize(int x, int y, int bpp);

    __fastcall TDDForm(TComponent* Owner);
    __fastcall ~TDDForm(void);

    __property bool DDPaused = {read=FAppPaused,write=SetAppPaused};

__published:
    __property System::AnsiString DDPalFileName = {read=FPalFileName, write=FPalFileName, nodefault};
    __property unsigned DDColorKeyRGB = {read=FColorKeyRGB, write=FColorKeyRGB, default=0};
    __property int DDWidth = {read=FDDWidth, write=SetDDWidth };
    __property int DDHeight = {read=FDDHeight, write=SetDDHeight };
    __property int DDBPP = {read=FDDBPP, write=SetDDBPP };
    __property bool DDActive = {read=FDDActive, write=SetDDActive};
    __property bool DDWindowed = {read=FAppWindowed,write=FAppWindowed};
    __property bool DDFullScreenClip = {read=FFullScreenClip,write=SetFullScreenClip};
    __property TNotifyEvent OnDDActiveChange = {read=FOnActiveChange, write=FOnActiveChange};
    __property TNotifyEvent OnDDUpdateDisplay = {read=FOnUpdateDisplay, write=FOnUpdateDisplay};
    __property TNotifyEvent OnDDChangeSize = {read=FOnChangeSize, write=FOnChangeSize};
    __property TNotifyEvent OnDDChangePalette = {read=FOnChangePalette, write=FOnChangePalette};
};

//---------------------------------------------------------------------------
class TDDThread : public TThread
{
private:
    TDDForm *DDForm;

protected:
    void __fastcall Execute();

public:
    __fastcall TDDThread(TDDForm *Owner) : TThread(true) { DDForm = Owner; };
};

//---------------------------------------------------------------------------
// extern PACKAGE TDDForm *DDForm;

//---------------------------------------------------------------------------
#endif
