//---------------------------------------------------------------------------
// Copyright (C) 2000 by Dmitriy Vassiliev
// SlyDiman@mail.ru
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <stdlib.h>
#include <math.h>
                                             
#include "DDSprite.h"

#pragma package(smart_init)
#pragma resource "*.res"

#define min(a,b)  (((a) < (b)) ? (a) : (b))
#define max(a,b)  (((a) > (b)) ? (a) : (b)) 

//---------------------------- TDDSpriteList -----------------------------

#define SListIndexError   "List index out of bounds (%d)"

TDDSprite *  __fastcall TDDSpriteList::Get(int Index)
{
  if( Index < 0 ||  Index >= Count )  Error( __classid(TDDSpriteList), SListIndexError, Index );
  return( (TDDSprite *)List[Index] );
}

void __fastcall TDDSpriteList::Put( int Index, TDDSprite * Item)
{
  if( Index < 0 || Index >= Count )  Error( __classid(TDDSpriteList), SListIndexError, Index );
  List[Index] = (void *)Item;
}

void __fastcall TDDSpriteList::ClearItems(void)
{
  for( int i=0; i < Count; i++ ){
    delete Items[i];
  }
  Clear();
}

void __fastcall TDDSpriteList::QuickSort( TListSortCompare SCompare )
{
  int l = 0;
  int r = Count-1;
  bool ch;
  do{
    ch = false;
    for( int i=l, j=l+1; i<r; i++, j++ ){
      if( SCompare( List[i], List[j] ) > 0 ){
        void *t = List[i];
        List[i] = List[j];
        List[j] = t;
        ch = true;
      }
    }
    --r;
    if( !ch )  return;
    ch = false;
    for( int i=r, j=r-1; i>l; i--, j-- ){
      if( SCompare( List[j], List[i] ) > 0 ){
        void *t = List[i];
        List[i] = List[j];
        List[j] = t;
        ch = true;
      }
    }
    ++l;
    if( !ch )  return;
  }while( l < r );
}



/*
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(TDDSprite *)
{
    new TDDSprite(NULL);
}
*/

//---------------------------------------------------------------------------
namespace Ddsprite
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TDDSprite)};
        RegisterComponents("SlyDiman", classes, 0);
    }
}

//---------------------------------------------------------------------------
__fastcall TDDSprite::TDDSprite(TComponent* Owner)
    : TComponent(Owner)
{
  if( !Owner->ClassParent()->ClassNameIs("TDDForm") )
    throw Exception ("TDDSprite: Owner Form is not TDDForm");

  DDForm = (TDDForm *)Owner;
  FVisible = true;
  FirstFrame = true;
  SecondFrame = false;

  FAutoBounce = false;

  LastTickCount = 0;

  VelocityX = 0;
  VelocityY = 0;
  PosX = 0;
  PosY = 0;

  FStart = 0;
  FStop = 0;
  FOffset = 0;
  FInterval = defAnimationInterval;
  FStartTick = 0;
  FLoop = true;
  FAutoHide = true;
  FMirror = false;

  FFramesPerLine = 1;
  FFirstFrameX = 0;
  FFirstFrameY = 0;
  FFrameW = 0;
  FFrameH = 0;

  FLeft = 0;
  FTop = 0;
  FWidth = 0;
  FHeight = 0;

  FFrames = NULL;

  FAlpha = 128;

  FLimPosX1 = 0;
  FLimPosX2 = 0;
  FLimPosY1 = 0;
  FLimPosY2 = 0;

  FLimMovX1 = 0;
  FLimMovX2 = 0;
  FLimMovY1 = 0;
  FLimMovY2 = 0;
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetVisible( bool Value )
{
  FVisible = Value;
  FirstFrame = true;
  SecondFrame = false;

//  LoopTime = (FStop - FStart + 1) * FInterval;
}

//---------------------------------------------------------------------------
int __fastcall TDDSprite::GetX(void)
{
  short Result = FX.Int;
  if( FX.Frac > 0x7FFF ) Result++;
  return( Result );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetX( int Value )
{
  FX.Int = (short)Value;
  FX.Frac = 0;
}

//---------------------------------------------------------------------------
int __fastcall TDDSprite::GetY(void)
{
  short Result = FY.Int;
  if( FY.Frac > 0x7FFF )  Result++;
  return( Result );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetY( int Value )
{
  FY.Int = (short)Value;
  FY.Frac = 0;
}

//---------------------------------------------------------------------------
double __fastcall TDDSprite::GetVelocityX(void)
{
  return( FVelocityX.Value / CoordMultiplier );
}


//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetVelocityX( double Value )
{
  FVelocityX.Value = (long)( Value * CoordMultiplier );
}

//---------------------------------------------------------------------------
double __fastcall TDDSprite::GetVelocityY(void)
{
  return( FVelocityY.Value / CoordMultiplier );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetVelocityY( double Value )
{
  FVelocityY.Value = (long)( Value * CoordMultiplier );
}

//---------------------------------------------------------------------------
double __fastcall TDDSprite::GetAccelerationX(void)
{
  return( FAccelerationX.Value / CoordMultiplier );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetAccelerationX( double Value )
{
  FAccelerationX.Value = (long)( Value * CoordMultiplier );
}

//---------------------------------------------------------------------------
double __fastcall TDDSprite::GetAccelerationY(void)
{
  return( FAccelerationY.Value / CoordMultiplier );
}


//---------------------------------------------------------------------------
void __fastcall TDDSprite::SetAccelerationY( double Value )
{
  FAccelerationY.Value = (long)( Value * CoordMultiplier );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::GetFrameNumber( long Ticks, unsigned &FrameNumber )
{
  // calculate the default next frame number
  FrameNumber = Start;

  if( Stop > Start ){
    if( Mirror ){
      unsigned int d = FStop - FStart;
      FrameNumber += ((Ticks - FStartTick) / FInterval + FOffset) % (d * 2);
      if( FrameNumber > d )
        FrameNumber = d * 2 - FrameNumber;
    }else{
      FrameNumber += ((Ticks - FStartTick) / FInterval + FOffset) % (FStop - FStart + 1);
    }

    if( !Loop ){
      if( FirstFrame ){
        FStartTick = Ticks;
        FrameNumber == Start;
        FirstFrame = false;
      }else{
        if( SecondFrame ){
          if( FrameNumber == Start ){
            if( AutoHide )  Visible = false;
            FrameNumber = FStop;
          }
        }else{
          if( FrameNumber > Start ){
            SecondFrame = true;
          }
        }
      }
    }

  }
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::GetFrameRect( long Ticks, RECT &ARect )
{
  unsigned ACurrentFrame;
  GetFrameNumber( Ticks, ACurrentFrame );

  ARect = Bounds( FirstFrameX + (ACurrentFrame % FramesPerLine )*FrameWidth  + ShowLeft,
                  FirstFrameY + (ACurrentFrame / FramesPerLine )*FrameHeight + ShowTop,
                  ShowWidth,
                  ShowHeight );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::GetFrameRectWH( long Ticks, RECT &ARect )
{
  unsigned ACurrentFrame;
  GetFrameNumber( Ticks, ACurrentFrame );

  ARect = Rect( FirstFrameX + (ACurrentFrame % FramesPerLine )*FrameWidth  + ShowLeft,
                FirstFrameY + (ACurrentFrame / FramesPerLine )*FrameHeight + ShowTop,
                ShowWidth,
                ShowHeight );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::GetShowRect( RECT &ARect )
{
  ARect = Bounds( PosX, PosY, ShowWidth, ShowHeight );
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::CheckBounce( void )
{
  // default CheckLimits calls this to make sprite bounce around
  if( (LimMovX1 < LimMovX2) && (LimMovY1 < LimMovY2) ){
    if( PosX + LimPosX1 <= LimMovX1 ){
      VelocityX = ABS( VelocityX );
    }else if( PosX + LimPosX2 >= LimMovX2 ){
      VelocityX = -ABS( VelocityX );
    }
    if( PosY + LimPosY1 <= LimMovY1 ){
      VelocityY = ABS( VelocityY );
    }else if( PosY + LimPosY2 >= LimMovY2 ){
      VelocityY = -ABS( VelocityY );
/*
      if( LastVelocityY >= -0.2 ){
        VelocityX = 0.4 - random(9) / 10;
        LastVelocityY = - (random(17) / 10) + 0.08;
      }else{
        LastVelocityY = (LastVelocityY * 9) / 10;
      }
      VelocityY = LastVelocityY;
*/
    }
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::CheckLimits( void )
{
// this routine is called by Move after the sprite has been moved,
// and can check and alter position, velocity or acceleration as required.

  bool DoDefaultProcessing = true;
  if( FOnCheckLimits ){
    DoDefaultProcessing = false;
    FOnCheckLimits( this, DoDefaultProcessing );
  }

  if( DoDefaultProcessing ){
    if( (LimMovX1 < LimMovX2) && (LimMovY1 < LimMovY2) ){
      if( PosX + LimPosX1 < LimMovX1 ){
        PosX = LimMovX1 - LimPosX1;
      }else if( PosX + LimPosX2 > LimMovX2 ){
        PosX = LimMovX2 - LimPosX2;
      }
      if( PosY + LimPosY1 < LimMovY1 ){
        PosY = LimMovY1 - LimPosY1;
      }else if( PosY + LimPosY2 > LimMovY2 ){
        PosY = LimMovY2 - LimPosY2;
      }
    }
    if( AutoBounce )  CheckBounce();
  }

  if( FOnCheckedLimits )  FOnCheckedLimits(this);
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::Move( long Ticks )
{
  int TicksSinceLastFrame = 0;
  if( LastTickCount != 0 )  TicksSinceLastFrame = Ticks - LastTickCount;
  LastTickCount = Ticks;
  if( TicksSinceLastFrame > MaxJump )  TicksSinceLastFrame = MaxJump;

  FX.Value += (FVelocityX.Value * TicksSinceLastFrame) / 1000;
  FY.Value += (FVelocityY.Value * TicksSinceLastFrame) / 1000;
  FVelocityX.Value += (FAccelerationX.Value * TicksSinceLastFrame) / 1000;
  FVelocityY.Value += (FAccelerationY.Value * TicksSinceLastFrame) / 1000;
  CheckLimits();
}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::Show( IDirectDrawSurface7 *pDDSurface, unsigned long Ticks, unsigned long Flags )
{
  if( !Visible )  return;
//  Move( Ticks );

  TRect SRect;
  GetShowRect( SRect );

  RECT FRect;
  DDSURFACEDESC2 SurfaceDesc;

 if( Flags & DDBLT_ALPHASRCCONSTOVERRIDE ){
    ZeroMemory(&SurfaceDesc, sizeof(SurfaceDesc));
    SurfaceDesc.dwSize = sizeof(SurfaceDesc);

    GetFrameRectWH( Ticks, FRect );
    pDDSurface->Lock(NULL, &SurfaceDesc, DDLOCK_WAIT, NULL);
    Frames->Lock();
    DDForm->AlphaEx->AlphaBlt( SurfaceDesc, SRect.left, SRect.top, Frames->SurfaceDesc, &FRect, (BYTE)Alpha );
    Frames->Unlock();
    pDDSurface->Unlock(NULL);

  }else if( Flags & DDBLT_ALPHASRC ){
    ZeroMemory(&SurfaceDesc, sizeof(SurfaceDesc));
    SurfaceDesc.dwSize = sizeof(SurfaceDesc);

    GetFrameRectWH( Ticks, FRect );
    pDDSurface->Lock(NULL, &SurfaceDesc, DDLOCK_WAIT, NULL);
    Frames->Lock();
    DDForm->AlphaEx->AlphaBltFire( SurfaceDesc, SRect.left, SRect.top, Frames->SurfaceDesc, &FRect );
    Frames->Unlock();
    pDDSurface->Unlock(NULL);

  }else if( Flags & DDBLT_ALPHASRCNEG ){
    ZeroMemory(&SurfaceDesc, sizeof(SurfaceDesc));
    SurfaceDesc.dwSize = sizeof(SurfaceDesc);

    GetFrameRectWH( Ticks, FRect );
    pDDSurface->Lock(NULL, &SurfaceDesc, DDLOCK_WAIT, NULL);
    Frames->Lock();
    DDForm->AlphaEx->AlphaBltGlass( SurfaceDesc, SRect.left, SRect.top, Frames->SurfaceDesc, &FRect );
    Frames->Unlock();
    pDDSurface->Unlock(NULL);

  }else{
    GetFrameRect( Ticks, FRect );
    pDDSurface->BltFast( SRect.left, SRect.top, Frames->pDDSurface, &FRect, Flags );
  }

}

//---------------------------------------------------------------------------
void __fastcall TDDSprite::OptimizeShowSize(void)
{
  Frames->Lock();

  DWORD ColorKey = *(DWORD *)Frames->SurfaceDesc.lpSurface;
  if( Frames->SurfaceDesc.ddpfPixelFormat.dwRGBBitCount < 32 )
      ColorKey &= (1 << Frames->SurfaceDesc.ddpfPixelFormat.dwRGBBitCount) - 1;

  unsigned int left = FrameWidth / 2;
  unsigned int right = left;

  unsigned int top = FrameHeight / 2;
  unsigned int bottom = top;

  for( int fy=0; fy < Frames->Height; fy+= FrameHeight ){
    for( int fx=0; fx < Frames->Width; fx+= FrameWidth ){

      for( unsigned int y=0; y < FrameHeight; y++ ){
        for( unsigned int x=0; x < FrameWidth; x++ ){

          switch( Frames->SurfaceDesc.ddpfPixelFormat.dwRGBBitCount ){
          case 16:
          {
            WORD *d = (WORD *)Frames->SurfaceDesc.lpSurface;
            DWORD c = d[ (fy+y)*Frames->Width + (fx+x) ];
            if( c != ColorKey ){
              left   = min( left,   x );
              right  = max( right,  x );
              top    = min( top,    y );
              bottom = max( bottom, y );
            }
          }
          break;
          case 32:
          {
            DWORD *d = (DWORD *)Frames->SurfaceDesc.lpSurface;
            DWORD c = d[ (fy+y)*Frames->Width + (fx+x) ];
            if( c != ColorKey ){
              left   = min( left,   x );
              right  = max( right,  x );
              top    = min( top,    y );
              bottom = max( bottom, y );
            }
          }
          break;
          }

        }
      }

    }
  }

  ShowLeft = left;
  ShowTop  = top;

  ShowWidth  = right - left + 1;
  ShowHeight = bottom - top + 1;

  Frames->Unlock();
}

//---------------------------------------------------------------------------



