//---------------------------------------------------------------------------
// Copyright (C) 1999 by Dmitriy Vassiliev
// SlyDiman@mail.ru
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#ifndef DDSpriteH
#define DDSpriteH

//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Forms.hpp>

#include "DDSurfac.h"

//---------------------------------------------------------------------------
#define  defAnimationInterval  (20)   // milliseconds
#define  CoordMultiplier       (65536.0000000)

#define MaxJump  500
#define  ABS(x)    (((x) < 0) ? -(x):(x))

typedef union {
  long Value;
  struct{
    unsigned short Frac;
    short Int;
  };
} _TCoord;

//---------------------------------------------------------------------------
typedef void __fastcall (__closure *TCheckLimitsEvent)( System::TObject *Sender,
                           bool &DoDefaultProcessing );

//---------------------------------------------------------------------------
class PACKAGE TDDSprite : public TComponent
{
protected:
    bool FirstFrame;
    bool SecondFrame;

    int FAlpha;
    bool FVisible;
    bool FAutoBounce;

    // position and size

    _TCoord FX;
    _TCoord FY;

    _TCoord FVelocityX;
    _TCoord FVelocityY;

    _TCoord FAccelerationX;
    _TCoord FAccelerationY;

    // limits
    int FLimPosX1;
    int FLimPosX2;
    int FLimPosY1;
    int FLimPosY2;

    int FLimMovX1;
    int FLimMovX2;
    int FLimMovY1;
    int FLimMovY2;

    bool     FLoop;
    bool     FAutoHide;
    bool     FMirror;
    unsigned FStartTick;
    unsigned FInterval;

    unsigned FStart;
    unsigned FStop;
    unsigned FOffset;

    unsigned FFramesPerLine;
    unsigned FFirstFrameX;
    unsigned FFirstFrameY;
    unsigned FFrameW;
    unsigned FFrameH;

    unsigned FWidth;
    unsigned FHeight;
    unsigned FLeft;
    unsigned FTop;

    unsigned LastTickCount;

    TDDForm *DDForm;
    TDDSurface *FFrames;

    // events
    TCheckLimitsEvent FOnCheckLimits;
    Classes::TNotifyEvent FOnCheckedLimits;

    virtual int    __fastcall GetX(void);
    virtual void   __fastcall SetX( int Value);
    virtual int    __fastcall GetY(void);
    virtual void   __fastcall SetY( int Value );
    virtual double __fastcall GetVelocityX(void);
    virtual void   __fastcall SetVelocityX( double Value );
    virtual double __fastcall GetVelocityY(void);
    virtual void   __fastcall SetVelocityY( double Value );
    virtual double __fastcall GetAccelerationX(void);
    virtual void   __fastcall SetAccelerationX( double Value );
    virtual double __fastcall GetAccelerationY(void);
    virtual void   __fastcall SetAccelerationY( double Value );
    virtual void   __fastcall SetVisible( bool Value );
    void __fastcall GetShowRect( RECT &ARect );

public:
    unsigned ID;

    unsigned User1;
    unsigned User2;
    unsigned User3;
    unsigned User4;
    unsigned short Keys[10];

    double LastVelocityX;   // meant to aid 'bounce damping' type calcs
    double LastVelocityY;
    virtual void __fastcall GetFrameNumber( long Ticks, unsigned &FrameNumber );
    void __fastcall GetFrameRect( long Ticks, RECT &ARect );
    void __fastcall GetFrameRectWH( long Ticks, RECT &ARect );
    virtual void __fastcall CheckBounce(void);
    virtual void __fastcall CheckLimits(void);
    virtual void __fastcall Move( long Ticks );
    void __fastcall OptimizeShowSize(void);
    void __fastcall Show( IDirectDrawSurface7 *pDDSurface, unsigned long Ticks, unsigned long Flags );
    __fastcall TDDSprite(TComponent* Owner);

__published:
    __property int LimPosX1 = {read=FLimPosX1, write=FLimPosX1};
    __property int LimPosX2 = {read=FLimPosX2, write=FLimPosX2};
    __property int LimPosY1 = {read=FLimPosY1, write=FLimPosY1};
    __property int LimPosY2 = {read=FLimPosY2, write=FLimPosY2};

    __property int LimMovX1 = {read=FLimMovX1, write=FLimMovX1};
    __property int LimMovX2 = {read=FLimMovX2, write=FLimMovX2};
    __property int LimMovY1 = {read=FLimMovY1, write=FLimMovY1};
    __property int LimMovY2 = {read=FLimMovY2, write=FLimMovY2};

    __property int   Alpha = {read=FAlpha, write=FAlpha};
    __property int PosX = {read=GetX, write=SetX};
    __property int PosY = {read=GetY, write=SetY};

    __property unsigned ShowWidth  = {read=FWidth, write=FWidth};
    __property unsigned ShowHeight = {read=FHeight, write=FHeight};
    __property unsigned ShowLeft = {read=FLeft, write=FLeft};
    __property unsigned ShowTop = {read=FTop, write=FTop};

    __property unsigned FrameWidth  = {read=FFrameW, write=FFrameW};
    __property unsigned FrameHeight = {read=FFrameH, write=FFrameH};
    __property unsigned FramesPerLine = {read=FFramesPerLine, write=FFramesPerLine};
    __property unsigned FirstFrameX = {read=FFirstFrameX, write=FFirstFrameX};
    __property unsigned FirstFrameY = {read=FFirstFrameY, write=FFirstFrameY};

    __property double VelocityX = {read=GetVelocityX, write=SetVelocityX};
    __property double VelocityY = {read=GetVelocityY, write=SetVelocityY};

    __property double AccelerationX = {read=GetAccelerationX, write=SetAccelerationX};
    __property double AccelerationY = {read=GetAccelerationY, write=SetAccelerationY};

    __property unsigned StartTick = {read=FStartTick, write=FStartTick};
    __property unsigned Interval  = {read=FInterval,  write=FInterval};
    __property unsigned Start     = {read=FStart,     write=FStart};
    __property unsigned Stop      = {read=FStop,      write=FStop};
    __property unsigned Offset    = {read=FOffset,    write=FOffset};
    __property bool     Loop      = {read=FLoop,      write=FLoop, default=true};
    __property bool     AutoHide  = {read=FAutoHide,  write=FAutoHide, default=true};
    __property bool     Mirror    = {read=FMirror,    write=FMirror, default=false};

    __property bool AutoBounce = {read=FAutoBounce, write=FAutoBounce, default=false};
    __property bool Visible = {read=FVisible, write=SetVisible, default=true};

    __property TDDSurface *Frames = {read=FFrames, write=FFrames};

    __property TCheckLimitsEvent OnCheckLimits = {read=FOnCheckLimits, write=FOnCheckLimits};
    __property Classes::TNotifyEvent OnCheckedLimits = {read=FOnCheckedLimits, write=FOnCheckedLimits};
};

//---------------------------------------------------------------------------
typedef TDDSprite * TDDSpriteListItem[134217727];
typedef TDDSpriteListItem *PDDSpriteListItem;

class PACKAGE TDDSpriteList : public TList
{
protected:
	TDDSprite * __fastcall Get(int Index);
	void __fastcall Put(int Index, TDDSprite * Item);

public:
  void __fastcall QuickSort( TListSortCompare SCompare );
  void __fastcall ClearItems(void);
  __property TDDSprite * Items[int Index] = {read=Get, write=Put};
  __fastcall TDDSpriteList( void ) : TList() {}
};

//---------------------------------------------------------------------------
#endif
