//---------------------------------------------------------------------------
// Copyright (C) 1999 by Dmitriy Vassiliev
// SlyDiman@mail.ru
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "DDSurfac.h"
#include "DDUtil.h"

#pragma package(smart_init)
#pragma link "LoadPcx"
#pragma resource "*.res"


//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(TDDSurface *)
{
    new TDDSurface(NULL);
}


//---------------------------------------------------------------------------
namespace Ddsurfac
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TDDSurface)};
        RegisterComponents("SlyDiman", classes, 0);
    }
}

//---------------------------------------------------------------------------
__fastcall TDDSurface::TDDSurface(TComponent* Owner) : TComponent(Owner)
{
  FWidth  = 0;
  FHeight = 0;
  FLocked = false;
  FFileType = ddsPcx;
  FAutoColorKey = true;

  pDDSurface = NULL;
  pDDCanvas  = NULL;

  if( ComponentState.Contains(csDesigning) && !Owner->ClassParent()->ClassNameIs("TDDForm") )
    MessageBox(NULL,"Owner Form is not TDDForm", "TDDSurface: Warning", MB_OK );

  DDForm = (TDDForm *)Owner;
}

//---------------------------------------------------------------------------
__fastcall TDDSurface::~TDDSurface(void)
{
  Release();
}

//---------------------------------------------------------------------------
void __fastcall TDDSurface::Init(void)
{
  Release();

  if( !DDForm || !DDForm->pDDraw )
    throw Exception ("TDDSurface: DirectDraw Form not initialized");

  ZeroMemory(&SurfaceDesc, sizeof(SurfaceDesc));
  SurfaceDesc.dwSize = sizeof(SurfaceDesc);

  if( FileName.IsEmpty() ){
    SurfaceDesc.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    SurfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    SurfaceDesc.dwWidth = Width;
    SurfaceDesc.dwHeight = Height;

    if( DDForm->pDDraw->CreateSurface( &SurfaceDesc, &pDDSurface, NULL) != DD_OK )
      throw Exception("TDDSurface: Create Surface FAILED");

    if( pDDSurface->GetSurfaceDesc(&SurfaceDesc) != DD_OK )
      throw Exception("TDDSurface: Get Surface Desc FAILED");

  }else{
    switch( FileType ){
    case ddsBmp:  pDDSurface = DDLoadBitmap( DDForm->pDDraw, FileName.c_str(), 0, 0 );  break;
    case ddsPcx:  pDDSurface = DDLoadPcx(    DDForm->pDDraw, FileName,         0, 0 );  break;
    }

    if( pDDSurface == NULL ){
      AnsiString s = "Unable to load surface " + FileName;
      throw Exception(s);
    }

    if( pDDSurface->GetSurfaceDesc(&SurfaceDesc) != DD_OK )
      throw Exception("TDDSurface: Get Surface Desc FAILED");

    Width  = SurfaceDesc.dwWidth;
    Height = SurfaceDesc.dwHeight;
  }

  if( AutoColorKey ){
    Lock();
    if( Locked ){
      DWORD Color = *(DWORD *)SurfaceDesc.lpSurface;
      if( SurfaceDesc.ddpfPixelFormat.dwRGBBitCount < 32 )
        Color &= (1 << SurfaceDesc.ddpfPixelFormat.dwRGBBitCount) - 1;
      Unlock();

      DDCOLORKEY              ddck;
      ddck.dwColorSpaceLowValue  = Color;
      ddck.dwColorSpaceHighValue = Color;
      pDDSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
    }
  }else{
    DDSetColorKey( pDDSurface, DDForm->DDColorKeyRGB );
  }

  pDDCanvas = new TDDCanvas( pDDSurface );
  try{
    pDDCanvas->Font->Name = "Arial";
    pDDCanvas->Font->Size = 12;
    pDDCanvas->Font->Color = (TColor)0x101010;     // clBlack
    pDDCanvas->Pen->Color  = (TColor)0x000000;     // transparent
    pDDCanvas->Brush->Color = clBlack;
    pDDCanvas->Brush->Style = bsSolid;
/*
    if( FileName.IsEmpty() )
      pDDCanvas->FillRect( Rect(0,0,Width-1,Height-1) );
*/
  }
  __finally{
    pDDCanvas->Release();
  }

  if( FileName.IsEmpty() ){
    DDBLTFX     ddbltfx;

    // Use the blter to do a color fill to clear the back buffer
    ZeroMemory(&ddbltfx, sizeof(ddbltfx));
    ddbltfx.dwSize = sizeof(ddbltfx);
    ddbltfx.dwFillColor = 0;
    pDDSurface->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
  }

/*
    HDC         hDC;
    if (Surface->GetDC(&hDC) == DD_OK)
    {
        // ...
        Surface->ReleaseDC(hDC);
    }

*/

  if( OnInit )  OnInit(this);
  if( OnUpdate )  OnUpdate( this );
}

//---------------------------------------------------------------------------
void __fastcall TDDSurface::Release(void)
{
  Unlock();

  if( pDDCanvas ){
    delete pDDCanvas;
    pDDCanvas = NULL;
  }

  if( pDDSurface ){
    pDDSurface->Release();
    pDDSurface = NULL;
  }
}

//---------------------------------------------------------------------------
void __fastcall TDDSurface::Update(void)
{
  if( pDDSurface == NULL )  return;

  Unlock();

  if( !FileName.IsEmpty() ){
    switch( FileType ){
    case ddsBmp:  DDReLoadBitmap( pDDSurface, FileName.c_str() );  break;
    case ddsPcx:  DDReLoadPcx(    pDDSurface, FileName         );  break;
    }
  }

  if( OnUpdate )  OnUpdate( this );
}

//---------------------------------------------------------------------------
void __fastcall TDDSurface::Lock(void)
{
  if( pDDSurface == NULL || Locked )  return;

  FLocked = (pDDSurface->Lock(NULL, &SurfaceDesc, DDLOCK_WAIT, NULL) == DD_OK );
}

//---------------------------------------------------------------------------
void __fastcall TDDSurface::Unlock(void)
{
  if( pDDSurface == NULL || !Locked )  return;

  pDDSurface->Unlock(NULL);

  FLocked = false;
}

//---------------------------------------------------------------------------

