//---------------------------------------------------------------------------
// Copyright (C) 2000 by SlyDiman
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "LoadPcx.h"

//---------------------------------------------------------------------------
bool LoadPcxPalette256( const AnsiString FileName, TRGBPal *Pal )
{
  TFileStream *FStream = NULL;

  try{
    FStream = new TFileStream( FileName, fmOpenRead | fmShareDenyNone );
  }
  catch(...){
    return(false);
  }

  int Size = FStream->Size;
  if( Size <= 128 + 768 )  goto Err;
  FStream->Position = Size - 769;
  BYTE code;
  FStream->Read( &code, 1 );
  if( code != 0x0C )  goto Err;
  if( FStream->Read( Pal, 768 ) != 768 )  goto Err;
  if( FStream )  delete FStream;
  return( true );

Err:
  if( FStream )  delete FStream;
  return( false );
}

//---------------------------------------------------------------------------
bool DDReLoadPcxCustom( IDirectDrawSurface7 *pDDSurface, int xs, int ys,
                        const AnsiString FileName, int xp, int yp, int w, int h )
{
  TRGBPal Pal[256];
  BYTE *Buff = NULL;
  TFileStream *FStream = NULL;
  int BuffSize = 0xFFF0;
  int BuffPos = 0;
  int PixelCount = 0;
  BYTE PixelData;
  bool Lock = false;

  try{
    FStream = new TFileStream( FileName, fmOpenRead | fmShareDenyNone );
  }
  catch(...){
    return(false);
  }

  if( FStream->Size < 128 + 768 )  goto Err;

  TPcxHeader PcxH;
  FStream->Position = 0;
  if( FStream->Read( &PcxH, 128 ) != 128 )  goto Err;

  if( PcxH.Manuf != 10 || PcxH.Hard != 5 || PcxH.Encod != 1 )  goto Err;

  {
    int wp = PcxH.x2 - PcxH.x1 + 1;
    int hp = PcxH.y2 - PcxH.y1 + 1;
    if( w <= 0 || xp + w > wp )  w = wp - xp;
    if( h <= 0 || yp + h > hp )  h = hp - yp;
  }

  if( PcxH.BitsPerPlane != 8 )  goto Err;

  if( PcxH.NumPlanes == 1 ){
    if( !LoadPcxPalette256( FileName, Pal ) )  goto Err;
  }

//---------------------------------
  DDSURFACEDESC2 ddsd;
  ddsd.dwSize = sizeof(ddsd);

  if( pDDSurface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL) != DD_OK )  goto Err;

  Lock = true;

  if( xs + w > ddsd.dwWidth  )  w = ddsd.dwWidth  - xs;
  if( ys + h > ddsd.dwHeight )  h = ddsd.dwHeight - ys;

  Buff = new unsigned char[BuffSize];

  FStream->Position = 128;

  BuffSize = FStream->Read( Buff, BuffSize );
  if( BuffSize < 1 )  goto Err;

  for( int y=0; y < yp + h; y++){
    for( int Plane=0; Plane < PcxH.NumPlanes; Plane++){
      for( int x=0; x < PcxH.BytesPerLine; x++){

        if( PixelCount == 0 ){
          if( BuffPos == BuffSize ){
            BuffSize = FStream->Read( Buff, BuffSize );
            if( BuffSize < 1 )  goto Err;
            BuffPos = 0;
          }
          PixelData = Buff[ BuffPos++ ];
          if( PixelData > 0xC0 ){
            PixelCount = PixelData - 0xC1;
            if( BuffPos == BuffSize ){
              BuffSize = FStream->Read( Buff, BuffSize );
              if( BuffSize < 1 )  goto Err;
              BuffPos = 0;
            }
            PixelData = Buff[ BuffPos++ ];
          }
        }else{
          --PixelCount;
        }

        if( (x >= xp) && (x < xp + w) &&
            (y >= yp) && (y < yp + h) ){

          if( PcxH.NumPlanes == 1 ){       // 256 colors

            TRGBPal *p = &Pal[PixelData];
            switch( ddsd.ddpfPixelFormat.dwRGBBitCount ){
            case 32: {
              unsigned long *Data = (unsigned long *)ddsd.lpSurface;
              Data += (y - yp + ys) * ddsd.dwWidth + (x - xp + xs);
              *Data = ((DWORD)p->r << 16) | ((DWORD)p->g << 8) | p->b;
            }
            break;
            case 16: {
              unsigned short *Data = (unsigned short *)ddsd.lpSurface;
              Data += (y - yp + ys) * ddsd.dwWidth + (x - xp + xs);
              *Data = COLOR_RGB_TO_16(p->b,p->g,p->r);
            }
            break;
            }

          }else{                           // True colors

            switch( ddsd.ddpfPixelFormat.dwRGBBitCount ){
            case 32: {
              unsigned long *Data = (unsigned long *)ddsd.lpSurface;
              Data += (y - yp + ys) * ddsd.dwWidth + (x - xp + xs);
              switch( Plane ){
              case 0:
                *Data =  (DWORD)PixelData << 16;
              break;
              case 1:
                *Data |= (DWORD)PixelData << 8;
              break;
              case 2:
                *Data |= (DWORD)PixelData;
              break;
              }
            }
            break;
            case 16: {
              unsigned short *Data = (unsigned short *)ddsd.lpSurface;
              Data += (y - yp + ys) * ddsd.dwWidth + (x - xp + xs);
              switch( Plane ){
              case 0:
                *Data  = (WORD)(PixelData & 0xF8) << 8;
              break;
              case 1:
                *Data |= (WORD)(PixelData & 0xFC) << 3;
              break;
              case 2:
                *Data |= (WORD)(PixelData & 0xF8) >> 3;
              break;
              }
            }
            break;
            }

          }
        }

      }
    }
  }

  pDDSurface->Unlock(NULL);

  if( Buff )  delete Buff;
  if( FStream )  delete FStream;
  return( true );

Err:
  if( Lock )  pDDSurface->Unlock(NULL);

  if( Buff )  delete Buff;
  if( FStream )  delete FStream;
  return( false );
}

//---------------------------------------------------------------------------
bool DDReLoadPcx( IDirectDrawSurface7 *pDDSurface, const AnsiString FileName )
{
  return( DDReLoadPcxCustom( pDDSurface, 0, 0, FileName, 0, 0, 0, 0 ) );
}

//-----------------------------------------------------------------------------
extern "C" IDirectDrawSurface7 * DDLoadPcx( IDirectDraw7 * pdd, const AnsiString FileName, int w, int h )
{
  if( w == 0 || h == 0 ){
    TFileStream *FStream = NULL;

    try{
      FStream = new TFileStream( FileName, fmOpenRead | fmShareDenyNone );
    }
    catch(...){
      return(NULL);
    }

    if( FStream->Size < 128 + 768 ){
      delete FStream;
      return(NULL);
    }

    TPcxHeader PcxH;
    FStream->Position = 0;
    if( FStream->Read( &PcxH, 128 ) != 128 ){
      delete FStream;
      return(NULL);
    }

    if( PcxH.Manuf != 10 || PcxH.Hard != 5 || PcxH.Encod != 1 ){
      delete FStream;
      return(NULL);
    }

    delete FStream;

    if( w == 0 )  w = PcxH.x2 - PcxH.x1 + 1;
    if( h == 0 )  h = PcxH.y2 - PcxH.y1 + 1;
  }

//-----------------
  IDirectDrawSurface7    *pdds;
  DDSURFACEDESC2          ddsd;
  ZeroMemory(&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
  ddsd.dwWidth  = w;
  ddsd.dwHeight = h;
  if( pdd->CreateSurface(&ddsd, &pdds, NULL) != DD_OK )  return NULL;

  DDReLoadPcxCustom( pdds, 0, 0, FileName, 0, 0, w, h );

  return pdds;
}

//---------------------------------------------------------------------------
bool LoadPcxBitmap( Graphics::TBitmap *Bitmap, int xs, int ys,
                        const AnsiString FileName, int xp, int yp, int w, int h )
{
  TRGBPal Pal[256];
  BYTE *Buff = NULL;
  TFileStream *FStream = NULL;
  int BuffSize = 0xFFF0;
  int BuffPos = 0;
  int PixelCount = 0;
  BYTE PixelData;

  try{
    FStream = new TFileStream( FileName, fmOpenRead | fmShareDenyNone );
  }
  catch(...){
    return(false);
  }

  if( FStream->Size < 128 + 768 )  goto Err;

  TPcxHeader PcxH;
  FStream->Position = 0;
  if( FStream->Read( &PcxH, 128 ) != 128 )  goto Err;

  if( PcxH.Manuf != 10 || PcxH.Hard != 5 || PcxH.Encod != 1 )  goto Err;

  {
    int wp = PcxH.x2 - PcxH.x1 + 1;
    int hp = PcxH.y2 - PcxH.y1 + 1;
    if( w <= 0 || xp + w > wp )  w = wp - xp;
    if( h <= 0 || yp + h > hp )  h = hp - yp;
  }

  if( PcxH.BitsPerPlane != 8 )  goto Err;

  if( PcxH.NumPlanes == 1 ){
    if( !LoadPcxPalette256( FileName, Pal ) )  goto Err;
  }

  if( Bitmap->Width == 0 || Bitmap->Height == 0 ){
    Bitmap->Width  = xs + w;
    Bitmap->Height = ys + h;
  }

//---------------------------------
  if( xs + w > Bitmap->Width  )  w = Bitmap->Width  - xs;
  if( ys + h > Bitmap->Height )  h = Bitmap->Height - ys;

  Buff = new unsigned char[BuffSize];

  FStream->Position = 128;

  BuffSize = FStream->Read( Buff, BuffSize );
  if( BuffSize < 1 )  goto Err;

  for( int y=0; y < yp + h; y++){
    for( int Plane=0; Plane < PcxH.NumPlanes; Plane++){
      for( int x=0; x < PcxH.BytesPerLine; x++){

        if( PixelCount == 0 ){
          if( BuffPos == BuffSize ){
            BuffSize = FStream->Read( Buff, BuffSize );
            if( BuffSize < 1 )  goto Err;
            BuffPos = 0;
          }
          PixelData = Buff[ BuffPos++ ];
          if( PixelData > 0xC0 ){
            PixelCount = PixelData - 0xC1;
            if( BuffPos == BuffSize ){
              BuffSize = FStream->Read( Buff, BuffSize );
              if( BuffSize < 1 )  goto Err;
              BuffPos = 0;
            }
            PixelData = Buff[ BuffPos++ ];
          }
        }else{
          --PixelCount;
        }

        if( (x >= xp) && (x < xp + w) &&
            (y >= yp) && (y < yp + h) ){

          if( PcxH.NumPlanes == 1 ){       // 256 colors

            DWORD Pixel = *((DWORD *)(&Pal[PixelData])) & 0xFFFFFF;
            Bitmap->Canvas->Pixels[x - xp + xs][y - yp + ys] = Pixel;

          }else{                           // True colors

            DWORD Pixel = Bitmap->Canvas->Pixels[x - xp + xs][y - yp + ys];

            switch( Plane ){
            case 0:
              Pixel =  (DWORD)PixelData;
            break;
            case 1:
              Pixel |= (DWORD)PixelData << 8;
            break;
            case 2:
              Pixel |= (DWORD)PixelData << 16;
            break;
            }

            Bitmap->Canvas->Pixels[x - xp + xs][y - yp + ys] = Pixel;
          }
        }

      }
    }
  }

  if( Buff )  delete Buff;
  if( FStream )  delete FStream;
  return( true );

Err:
  if( Buff )  delete Buff;
  if( FStream )  delete FStream;
  return( false );
}

//-----------------------------------------------------------------------------
bool GetPcxSize( const AnsiString FileName, int &w, int &h )
{
    TFileStream *FStream = NULL;

    try{
      FStream = new TFileStream( FileName, fmOpenRead | fmShareDenyNone );
    }
    catch(...){
      return(false);
    }

    if( FStream->Size < 128 + 768 ){
      delete FStream;
      return(false);
    }

    TPcxHeader PcxH;
    FStream->Position = 0;
    if( FStream->Read( &PcxH, 128 ) != 128 ){
      delete FStream;
      return(false);
    }

    if( PcxH.Manuf != 10 || PcxH.Hard != 5 || PcxH.Encod != 1 ){
      delete FStream;
      return(false);
    }

    delete FStream;

    w = PcxH.x2 - PcxH.x1 + 1;
    h = PcxH.y2 - PcxH.y1 + 1;

    return( true );
}

//---------------------------------------------------------------------------




