//---------------------------------------------------------------------------
#ifndef LoadPcxH
#define LoadPcxH

#include <windows.h>
#include <DDraw.h>

#include <graphics.hpp>

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
typedef struct {
	BYTE Manuf;         // Always =10 for Paintbrush
	BYTE Hard;          // Version information
	BYTE Encod;         // Run-length encoding (=1)
	BYTE BitsPerPlane;  // Bits per pixel
	WORD x1;            // Picture dimensions (incl)
	WORD y1;
	WORD x2;
	WORD y2;
	WORD HRes;          // Display horiz resolution
	WORD VRes;          // Display vert  resolution
	BYTE Pal16[48];     // Pallete
	BYTE VMode;         // (ignored)
	BYTE NumPlanes;     // Number of planes (ver 2.5=0)
	WORD BytesPerLine;  // Bytes per line
	WORD ColorGray;     // Palette Info (1=col, 2=gray)
	WORD ScanHRes;      // Scanner resolution
	WORD ScanVRes;
	BYTE Extra[54];     // Extra space (filler)
} TPcxHeader;

//---------------------------------------------------------------------------
typedef struct{
  BYTE r;
  BYTE g;
  BYTE b;
} TRGBPal;

//---------------------------------------------------------------------------
#define COLOR_TRANSP(c1, c2, m)  (c1 += ((int)c2 - c1) * (int)m / 256)

#define COLOR_RGB_TO_15(r,g,b) ( (((WORD)(b) & 0xF8) << 7) | \
                                 (((WORD)(g) & 0xF8) << 2) | \
                                 (((WORD)(r) & 0xF8) >> 3) )

#define COLOR_RGB_TO_16(r,g,b) ( (((WORD)(b) & 0xF8) << 8) | \
                                 (((WORD)(g) & 0xFC) << 3) | \
                                 (((WORD)(r) & 0xF8) >> 3) )

#define COLOR_RGB_TO_24(r,g,b) ( (((DWORD)(b) & 0xFF) << 16) | \
                                 (((DWORD)(g) & 0xFF) <<  8) | \
                                 (((DWORD)(r) & 0xFF)      ) )

#define COLOR_15_TO_16(c)   ( (((WORD)(c) & 0x7FE0) << 1) | ((WORD)(c) & 0x1F) )

#define COLOR_24_TO_16(c)      ( (((b) & 0xF80000) >> 8) | \
                                 (((g) & 0x00FC00) >> 5) | \
                                 (((r) & 0x0000F8) >> 3) )

#define COLOR_32_TO_16(rgb) ( ((WORD)(rgb >> 8) & 0xF800) | \
                              ((WORD)(rgb >> 5) & 0x07E0) | \
                              ((WORD)(rgb >> 3) & 0x001F) )

//---------------------------------------------------------------------------
extern bool DDReLoadPcxCustom( IDirectDrawSurface7 *pDDSurface, int xs, int ys,
                        const AnsiString FileName, int xp, int yp, int w, int h );

extern bool DDReLoadPcx( IDirectDrawSurface7 *pDDSurface, const AnsiString FileName );
extern "C" IDirectDrawSurface7 * DDLoadPcx( IDirectDraw7 * pdd, const AnsiString FileName, int w, int h );

bool LoadPcxBitmap( Graphics::TBitmap *Bitmap, int xs, int ys,
                        const AnsiString FileName, int xp, int yp, int w, int h );

bool GetPcxSize( const AnsiString FileName, int &w, int &h );

//---------------------------------------------------------------------------
#endif


